c 3D PPR UEL for ABAQUS Code (Compatible with 8 node Brick Elements)
c By: Daniel Spring
c October 3rd, 2012
c
c References
c 1) Seong Hyeok Song "Fracture of Asphalt Concrete: A Cohesive
c Zone Modeling Approach Considering Viscoelastic Effects." PhD Thesis,
c Department of Civil and Environmental Engineering, UIUC, 2006.
c
c 2) Kyoungsoo Park. "Concrete Fracture Mechanics and Size Effect Using
c a Specialized Cohesive Zone Model" MS Thesis, Department of Civil and
c Environmental Engineering, UIUC, 2005.
c
c 3) "Cohesive fracture model for functionally graded fiber reinforced
c concrete" K. Park, G.H. Paulino, J. Roesler. Cement and Concrete
c Research. Vol 40, No. 6, pp. 956-965, 2010.
c =====================================================================
      SUBROUTINE UEL (RHS, AMATRX, SVARS, ENERGY, NDOFEL, NRHS, NSVARS,
     1 PROPS, NPROPS, COORDS, MCRD, NNODE, U, DU, V, A, JTYPE, TIME,
     2 DTIME, KSTEP, KINC, JELEM, PARAMS, NDLOAD, JDLTYP, ADLMAG,
     3 PREDEF, NPREDF, LFLAGS, MLVARX, DDLMAG, MDLOAD, PNEWDT, JPROPS,
     4 NJPROP, PERIOD)
c
      INCLUDE 'ABA_PARAM.INC'
c     
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),COORDS(MCRD,NNODE),U(NDOFEL),
     2 DU(MLVARX,*),V(NDOFEL),A(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)
c
      DIMENSION ds1(4),ds2(4),dn(4),Trac(MCRD,NRHS),
     1 Trac_Jacob(MCRD,MCRD),R(MCRD,MCRD),coord_l(MCRD,NNODE),
     2 GP_coord(2),sf(4),B(MCRD,NDOFEL),co_de_m(3,4),
     3 B_t(NDOFEL,MCRD), Transformation_M(NDOFEL,NDOFEL),
     4 Transformation_M_T(NDOFEL,NDOFEL),temp1(MCRD,NDOFEL)
c
      DIMENSION stiff_l(NDOFEL,NDOFEL),temp2(NDOFEL,NDOFEL),
     1 stiff_g(NDOFEL,NDOFEL),residual_l(NDOFEL,NRHS),
     2 residual_g(NDOFEL,NRHS),aJacob_M(2,3),delu_loc_gp(mcrd),
     3 co_de(mcrd,nnode)
c
      DOUBLE PRECISION G_fn, G_ft, f_tn, f_tt, alpha, beta, rn, rt,
     1 alphaV, betaV, alphaP, p_m, p_n, deln, delt, tmax, pmax, opn, 
     2 opt
c
c Define Inputs========================================================
c 
      G_fn=PROPS(1)
      G_ft=PROPS(2)
      f_tn=PROPS(3)
      f_tt=PROPS(4)
      alpha=PROPS(5)
      beta=PROPS(6)
      rn=PROPS(7)
      rt=PROPS(8)
      alphaV=1
      betaV=1
      alphaP=1
      GP_n=4d0
c
c Initialize Matrices and Vectors======================================
c 
      call k_vector_zero(ds1,4)
      call k_vector_zero(ds2,4)
      call k_vector_zero(dn,4)
      call k_matrix_zero(Trac,mcrd,nrhs)
      call k_matrix_zero(Trac_Jacob,mcrd,mcrd)
      call k_matrix_zero(R,mcrd,mcrd)
      call k_matrix_zero(coord_l,mcrd,nnode)
      call k_vector_zero(GP_coord,2)
      call k_vector_zero(sf,4)
      call k_matrix_zero(Transformation_M,ndofel,ndofel)
      call k_matrix_zero(Transformation_M_T,ndofel,ndofel)
      call k_matrix_zero(B,mcrd,ndofel)
      call k_matrix_zero(B_t,ndofel,mcrd)
      call k_matrix_zero(temp1,mcrd,ndofel)
      call k_matrix_zero(stiff_l,ndofel,ndofel)
      call k_matrix_zero(temp2,ndofel,ndofel)
      call k_matrix_zero(stiff_g,ndofel,ndofel)
      call k_matrix_zero(residual_l,ndofel,nrhs)
      call k_matrix_zero(residual_g,ndofel,nrhs)
      call k_matrix_zero(aJacob_M,2,3)
      call k_matrix_zero(rhs,ndofel,nrhs)
      call k_matrix_zero(amatrx,ndofel,ndofel)
      call k_matrix_zero(co_de,mcrd,nnode)
      a_Jacob=0.d0
c
c Determine Inputs to Cohesive Model===================================
c
      p_m=(alpha*(alpha-1.0)*rn**2.0)/(1.0-alpha*rn**2.0)
      p_n=(beta*(beta-1.0)*rt**2.0)/(1.0-beta*rt**2.0)
      deln=(G_fn/f_tn)*alpha*rn*(1.0-rn)**(alpha-1.0)
     & *((alpha/p_m)+1.0)*((alpha/p_m)*rn+1.0)**(p_m-1.0)
      delt=(G_ft/f_tt)*beta*rt*(1.0-rt)**(beta-1.0)
     & *((beta/p_n)+1.0)*((beta/p_n)*rt+1.0)**(p_n-1.0)
c
c Do local computations================================================
c 
      do i = 1, mcrd
         do j = 1, nnode
            co_de(i,j)=coords(i,j)+U(3.0*(j-1.0)+i)
         end do
      end do
c
c Do Calculations at Gauss Points======================================
c
      do i = 1, GP_n
c
      call k_matrix_zero(aJacob_M,2,3)
c
      gpt = i
c
      call k_local_coordinates(gpt,co_de,R,coord_l,Transformation_M,
     & Transformation_M_T,a_Jacob,aJacob_M,coords,u,ndofel,nnode,
     & mcrd, SVARS)
c
c Compute shear and normal local opening displacments==================
c 
      do j = 1, 4
         ds1(j)=coord_l(1,j+4)-coord_l(1,j)
         ds2(j)=coord_l(2,j+4)-coord_l(2,j)
         dn(j) =coord_l(3,j+4)-coord_l(3,j)
      end do
c
c Determine the values of the shape function at each Gauss Point
c
         call k_shape_fun(i,sf)
c
         call k_vector_zero(delu_loc_gp,mcrd)
c
c Determine shear and normal opening displamenets at Gauss points
c         
         do j = 1, 4
            delu_loc_gp(1)=delu_loc_gp(1)+ds1(j)*sf(j)
            delu_loc_gp(2)=delu_loc_gp(2)+ds2(j)*sf(j)
            delu_loc_gp(3)=delu_loc_gp(3)+ dn(j)*sf(j)
         end do
c
         opn=delu_loc_gp(3)
         opt=sqrt(delu_loc_gp(1)**2.0+delu_loc_gp(2)**2.0)
c
         if ((Svars(GP_n*(i-1.0)+1.0) .LT. opt) .AND.
     & (opt .GT. rt*delt)) then
            Svars(GP_n*(i-1.0)+1.0)=opt
         end if
         if ((Svars(GP_n*(i-1.0)+2.0) .LT. opn) .AND.
     & (opn .GT. rn*deln)) then
            Svars(GP_n*(i-1.0)+2.0)=opn
         end if
         tmax=Svars(GP_n*(i-1.0)+1.0)
         pmax=Svars(GP_n*(i-1.0)+2.0)
c
c Determine Traction vector and tangent modulus matrix
c 
         call k_cohesive_law(Trac,Trac_Jacob,G_fn,G_ft,deln,delt,
     & alpha,beta,p_m,p_n,pmax,tmax,alphaV,betaV,alphaP,delu_loc_gp,
     & mcrd, nrhs, SVARS)
c
c Determine B matrix and its transpose
c 
         call k_Bmatrix(sf,B,mcrd,ndofel)
c
         call k_matrix_transpose(B,B_t,mcrd,ndofel)
c
c Compute the stiffness matrix
c Local Stiffness = B_t * Trac_Jacob * B
c
         call k_matrix_multiply(Trac_Jacob,B,temp1,mcrd,mcrd,
     & ndofel)
         call k_matrix_multiply(B_t,temp1,stiff_l,ndofel,
     & mcrd,ndofel)
c
c Compute Global stiffness matrix 
c Global_K = T' * K * T
c
         call k_matrix_multiply(Transformation_M_T,stiff_l,
     & temp2,ndofel,ndofel,ndofel)
         call k_matrix_multiply(temp2,Transformation_M,stiff_g,
     & ndofel,ndofel,ndofel)
c
c Multiply Jacobian with the Global stiffness and add contribution
c from each Gauss Point
c
         call k_matrix_plus_scalar(amatrx,stiff_g,a_Jacob,
     & ndofel,ndofel)
c
c Compute the global residual vector
c Local_residual = B_t * Trac
c Global_residual = T' * Local_residual
c 
         call k_matrix_multiply(B_t,Trac,residual_l,ndofel,
     & mcrd,nrhs)
         call k_matrix_multiply(Transformation_M_T,residual_l,
     & residual_g,ndofel,ndofel,nrhs)
c
c Multiply the Global residual by the Jacobian and add the 
c contribution from each point
c   
         call k_matrix_plus_scalar(rhs,residual_g,a_Jacob,
     & ndofel,nrhs)
      end do
c
      return
      end
c======================================================================
c=============================SUBROUTINES==============================
c======================================================================
c
c Determine the strain-displacement (B) matrix
c
      subroutine k_Bmatrix(sf,B,mcrd,ndofel)
      INCLUDE 'ABA_PARAM.INC'
      dimension sf(4),B(mcrd,ndofel)
      B(1,1) =  sf(1)
      B(1,4) =  sf(2)
      B(1,7) =  sf(3)
      B(1,10)=  sf(4)
      B(1,13)= -sf(1)
      B(1,16)= -sf(2)
      B(1,19)= -sf(3)
      B(1,22)= -sf(4)
      B(2,2) =  sf(1)
      B(2,5) =  sf(2)
      B(2,8) =  sf(3)
      B(2,11)=  sf(4)
      B(2,14)= -sf(1)
      B(2,17)= -sf(2)
      B(2,20)= -sf(3)
      B(2,23)= -sf(4)
      B(3,3) =  sf(1)
      B(3,6) =  sf(2)
      B(3,9) =  sf(3)
      B(3,12)=  sf(4)
      B(3,15)= -sf(1)
      B(3,18)= -sf(2)
      B(3,21)= -sf(3)
      B(3,24)= -sf(4)
c
      return
      end
c======================================================================
      subroutine k_cohesive_law(T,T_d,G_fn,G_ft,deln,delt,
     & alpha,beta,p_m,p_n,pmax,tmax,alphaV,betaV,alphaP,delu,
     & mcrd, nrhs, SVARS)
      INCLUDE 'ABA_PARAM.INC'
      dimension T(mcrd,nrhs),T_d(mcrd,mcrd),delu(mcrd),SVARS(38)
       DOUBLE PRECISION G_fn, G_ft, f_tn, f_tt, alpha, beta,
     & alphaV, betaV, alphaP, p_m, p_n, deln, delt, tmax, pmax,
     & popn, popt, gam_n, gam_t, Tn, Tt, Dnn, Dnt, Dtt, T_d, T,
     & delu 
c
      popn=delu(3)
      popt=sqrt(delu(1)**2.0+delu(2)**2.0)
c
      call k_Mac(pM1,G_fn,G_ft)
      call k_Mac(pM2,G_ft,G_fn)
c
      if (G_fn .NE. G_ft) then
         gam_n=(-G_fn)**(pM1/(G_fn-G_ft))*(alpha/p_m)**p_m
         gam_t=(-G_ft)**(pM2/(G_ft-G_fn))*(beta/p_n)**p_n
      elseif (G_fn .EQ. G_ft) then
         gam_n=-G_fn*(alpha/p_m)**p_m
         gam_t=(beta/p_n)**p_n
      end if
c
c Pre-calculation of the normal cohesive traction Tn
c
      if (popn .LT. 0.0) then
         popn=0.0
      elseif ((popn .GE. deln) .OR. (popt .GE. delt)) then
         Tn = 0.0
      elseif (popn .GE. pmax) then
         Tn=(gam_n/deln)*(p_m*(1.0-(popn/deln))**alpha*((p_m/alpha)
     & +(popn/deln))**(p_m-1.0)-alpha*((1.0-(popn/deln))**(alpha-1.0))
     & *((p_m/alpha)+(popn/deln))**p_m)*(gam_t*(1.0-(popt/delt))
     & **beta*((p_n/beta)+(popt/delt))**p_n+pM2)
      else
         Tn=(gam_n/deln)*(p_m*(1.0-(pmax/deln))**alpha*((p_m/alpha)
     & +(pmax/deln))**(p_m-1.0)-alpha*((1.0-(pmax/deln))**(alpha-1.0))
     & *((p_m/alpha)+(pmax/deln))**p_m)*(gam_t*(1.0-(popt/delt))
     & **beta*((p_n/beta)+(popt/delt))**p_n+pM2)*popn/pmax
      end if
c
c Pre-calculation of the tangential cohesive traction, Tt
c
      if ((popn .GE. deln) .OR. (popt .GE. delt)) then
         Tt = 0.0
      elseif (popt .GE. tmax) then
         Tt=(gam_t/delt)*(p_n*(1.0-(popt/delt))**beta*((p_n/beta)
     & +(popt/delt))**(p_n-1.0)-beta*((1.0-(popt/delt))**(beta-1.0))
     & *((p_n/beta)+(popt/delt))**p_n)*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1) 
      else
         Tt=(gam_t/delt)*(p_n*(1.0-(tmax/delt))**beta*((p_n/beta)
     & +(tmax/delt))**(p_n-1.0)-beta*((1.0-(tmax/delt))**(beta-1.0))
     & *((p_n/beta)+(tmax/delt))**p_n)*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1)*popt/tmax     
      end if 
c
c Algortihm 1 
c
c Normal Cohesive Interaction
c (1) Contact
      if (delu(3) .LT. 0.0) then
         Dnn = -(gam_n/(deln**2))*(p_m/alpha)**(p_m-1.0)*(alpha+p_m)*
     & (gam_t*(p_n/beta)**p_n + pM2)
         Dnt = 0.0
         Tn = Dnn * delu(3)
      else if ((popn .LT. deln) .AND. (popt .LT. delt) 
     & .AND. (Tn .GE. -1.0E-5)) then
         Tn = Tn
c (2) Softening Condition
         if (popn .GE. pmax) then
         Dnn=(gam_n/(deln**2.0))*((p_m**2.0-p_m)*((1.0-(popn/deln))
     & **alpha)*((p_m/alpha)+(popn/deln))**(p_m-2.0)+(alpha**2.0-alpha)
     & *((1.0-(popn/deln))**(alpha-2.0))*((p_m/alpha)+(popn/deln))**p_m
     & -2.0*alpha*p_m*((1.0-(popn/deln))**(alpha-1.0))*((p_m/alpha)
     & +(popn/deln))**(p_m-1.0))*(gam_t*(1.0-(popt/delt))**beta
     &*((p_n/beta)+(popt/delt))**p_n+pM2)
         Dnt=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(popn/deln))**alpha)
     & *((p_m/alpha)+(popn/deln))**(p_m-1.0)-alpha*((1.0-(popn/deln))
     & **(alpha-1.0))*((p_m/alpha)+(popn/deln))**p_m)*(p_n
     & *((1.0-(popt/delt))**beta)*(((p_n/beta)+(popt/delt))**(p_n-1.0))
     & -beta*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)+(popt/delt))
     & **p_n)
c (3) Unloading/reloading condition
         else
         Dnn=(gam_n/deln)*(p_m*(1.0-(pmax/deln))**alpha*((p_m/alpha)
     & +(pmax/deln))**(p_m-1.0)-alpha*((1.0-(pmax/deln))**(alpha-1.0))
     & *((p_m/alpha)+(pmax/deln))**p_m)*(gam_t*(1.0-(popt/delt))
     & **beta*((p_n/beta)+(popt/delt))**p_n+pM2)/pmax
         Dnt=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(pmax/deln))**alpha)
     & *((p_m/alpha)+(pmax/deln))**(p_m-1.0)-alpha*((1.0-(pmax/deln))
     & **(alpha-1.0))*((p_m/alpha)+(pmax/deln))**p_m)*(p_n
     & *((1.0-(popt/delt))**beta)*(((p_n/beta)+(popt/delt))**(p_n-1.0))
     & -beta*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)+(popt/delt))
     & **p_n)*popn/pmax
         end if
c (4) Complete Failure
      else
         Tn = 0.0
         Dnn = 0.0
         Dnt = 0.0
      end if    
c
c Tangential Cohesive Interaction
c
      if ((popt .LT. delt) .AND. (popn .LT. deln) 
     & .AND. (Tt .GE. -1.0E-5)) then
        Tt = Tt
c (1) Softening Condition
        if (popt .GE. tmax) then
           Dtt=(gam_t/(delt**2.0))*((p_n**2.0-p_n)*((1.0-(popt/delt))
     & **beta)*((p_n/beta)+(popt/delt))**(p_n-2.0)+(beta**2.0-beta)
     & *((1.0-(popt/delt))**(beta-2.0))*((p_n/beta)+(popt/delt))**p_n
     & -2.0*beta*p_n*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)
     & +(popt/delt))**(p_n-1.0))*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1)
           Dtn=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(popn/deln))
     & **alpha)*((p_m/alpha)+(popn/deln))**(p_m-1.0)-alpha*((1.0
     & -(popn/deln))**(alpha-1.0))*((p_m/alpha)+(popn/deln))**p_m)*(p_n
     & *((1.0-(popt/delt))**beta)*(((p_n/beta)+(popt/delt))**(p_n-1.0))
     & -beta*((1.0-(popt/delt))**(beta-1.0))*((p_n/beta)+(popt/delt))
     & **p_n)
c (2) Unloading/reloading condition
        else
           Dtt=(gam_t/delt)*(p_n*(1.0-(tmax/delt))**beta*((p_n/beta)
     & +(tmax/delt))**(p_n-1.0)-beta*((1.0-(tmax/delt))**(beta-1.0))
     & *((p_n/beta)+(tmax/delt))**p_n)*(gam_n*(1.0-(popn/deln))
     & **alpha*((p_m/alpha)+(popn/deln))**p_m+pM1)/tmax 
           Dtn=(gam_n*gam_t/(deln*delt))*(p_m*((1.0-(popn/deln))
     & **alpha)*((p_m/alpha)+(popn/deln))**(p_m-1.0)-alpha*((1.0
     & -(popn/deln))**(alpha-1.0))*((p_m/alpha)+(popn/deln))**p_m)*(p_n
     & *((1.0-(tmax/delt))**beta)*(((p_n/beta)+(tmax/delt))**(p_n-1.0))
     & -beta*((1.0-(tmax/delt))**(beta-1.0))*((p_n/beta)+(tmax/delt))
     & **p_n)*popt/tmax
        end if
c (3) Complete failure condition
      else
        Tt = 0.0
        Dtt = 0.0
        Dtn = 0.0
      end if
      if (Dtn .NE. Dnt) then
        Dtn = 0.5*(Dtn + Dnt)
        Dnt = Dtn
      end if
c
      if (popt .EQ. 0.0) then
         T(1,1) = 0.0
         T(2,1) = 0.0
         T(3,1) = Tn
c
         T_d(1,1)=Dtt
         T_d(1,2)=0.0d00
         T_d(1,3)=0.0d00
         T_d(2,1)=0.0d00
         T_d(2,2)=Dtt
         T_d(2,3)=0.0d00
         T_d(3,1)=0.0d00
         T_d(3,2)=0.0d00
         T_d(3,3)=Dnn
      else
         T(1,1)=Tt*delu(1)/popt
         T(2,1)=Tt*delu(2)/popt
         T(3,1)=Tn
c
         T_d(1,1)=Dtt*(delu(1)/popt)**2.0+Tt*((delu(2)**2.0)/(popt
     &    **3.0))
         T_d(1,2)=Dtt*delu(1)*delu(2)/(popt**2.0)
     &    -Tt*delu(1)*delu(2)/(popt**3.0)
         T_d(1,3)=Dnt*delu(1)/popt
c
         T_d(2,1)=Dtt*delu(1)*delu(2)/(popt**2.0)
     &    -Tt*delu(1)*delu(2)/(popt**3.0)
         T_d(2,2)=Dtt*(delu(2)/popt)**2.0+Tt*((delu(1)**2.0)/(popt
     &    **3.0))
         T_d(2,3)=Dnt*delu(2)/popt
c
         T_d(3,1)=Dtn*delu(1)/popt
         T_d(3,2)=Dtn*delu(2)/popt
         T_d(3,3)=Dnn
      end if
c
      return
      end
c=====================================================================
      subroutine k_local_coordinates(gpt,co_de,R,coord_l,Transformation_M,
     & Transformation_M_T,a_Jacob,aJacob_M,coords,u,ndofel,nnode,
     & mcrd, SVARS)
      INCLUDE 'ABA_PARAM.INC'
      dimension R(mcrd,mcrd),coord_l(mcrd,nnode),aJacob_M(2,3),
     & Transformation_M(ndofel,ndofel),coords(mcrd,nnode),
     & Transformation_M_T(ndofel,ndofel),u(ndofel),
     & co_de(mcrd,nnode), co_de_m(3,4),SFD(2,4),SVARS(38)
       DOUBLE PRECISION x1, x2, x3, x4, y1, y2, y3, y4, y5, y6, z1, z2,
     & z3, z4
c
      call k_matrix_zero(co_de_m,3,4)
c
      do i = 1, 3
         co_de_m(i,1)=(co_de(i,1)+co_de(i,5))*0.5
         co_de_m(i,2)=(co_de(i,2)+co_de(i,6))*0.5
         co_de_m(i,3)=(co_de(i,3)+co_de(i,7))*0.5
         co_de_m(i,4)=(co_de(i,4)+co_de(i,8))*0.5
      end do
c
      x1=co_de_m(1,1)
      x2=co_de_m(1,2)
      x3=co_de_m(1,3)
      x4=co_de_m(1,4)
c
      y1=co_de_m(2,1)
      y2=co_de_m(2,2)
      y3=co_de_m(2,3)
      y4=co_de_m(2,4)
c
      z1=co_de_m(3,1)
      z2=co_de_m(3,2)
      z3=co_de_m(3,3)
      z4=co_de_m(3,4)
c
      if (gpt .eq. 1) then
         c_r=-sqrt(1.0d0/3.0d0)
         c_s=-sqrt(1.0d0/3.0d0)
      elseif (gpt .eq. 2) then
         c_r= sqrt(1.0d0/3.0d0)
         c_s=-sqrt(1.0d0/3.0d0)
      elseif (gpt .eq. 3) then
         c_r= sqrt(1.0d0/3.0d0)
         c_s= sqrt(1.0d0/3.0d0)
      elseif (gpt .eq. 4) then
         c_r=-sqrt(1.0d0/3.0d0)
         c_s= sqrt(1.0d0/3.0d0)
      end if
c
      SFD(1,1) =-0.25*(1-c_s)
      SFD(1,2) = 0.25*(1-c_s)
      SFD(1,3) = 0.25*(1+c_s)
      SFD(1,4) =-0.25*(1+c_s)
      SFD(2,1) =-0.25*(1-c_r)
      SFD(2,2) =-0.25*(1+c_r)
      SFD(2,3) = 0.25*(1+c_r)
      SFD(2,4) = 0.25*(1-c_r)
c
      do i = 1,2
         do j = 1,3
            do k =1, 4
               aJacob_M(i,j) = aJacob_M(i,j) + SFD(i,k)*co_de_m(j,k)
            end do
         end do
      end do
c
      dum1 = aJacob_M(1,2)*aJacob_M(2,3) - aJacob_M(1,3)*aJacob_M(2,2)
      dum2 = aJacob_M(1,3)*aJacob_M(2,1) - aJacob_M(1,1)*aJacob_M(2,3)
      dum3 = aJacob_M(1,1)*aJacob_M(2,2) - aJacob_M(1,2)*aJacob_M(2,1)
c
      a_Jacob = sqrt(dum1**2 + dum2**2 + dum3**2)
c
      R(3,1) = dum1/a_Jacob
      R(3,2) = dum2/a_Jacob
      R(3,3) = dum3/a_Jacob
c
      aLen=sqrt(aJacob_M(1,1)**2.0d00+aJacob_M(1,2)**2.0d00+aJacob_M(1,3)**2.0d00)
      R(1,1)=aJacob_M(1,1)/aLen
      R(1,2)=aJacob_M(1,2)/aLen
      R(1,3)=aJacob_M(1,3)/aLen
c
      aLen2a=R(3,2)*R(1,3)-R(3,3)*R(1,2)
      aLen2b=R(3,3)*R(1,1)-R(3,1)*R(1,3)
      aLen2c=R(3,1)*R(1,2)-R(3,2)*R(1,1)
c
      aLen2 = sqrt(aLen2a**2.0d00 + aLen2b**2.0d00 + aLen2c**2.0d00)
c
      R(2,1) = aLen2a/aLen2
      R(2,2) = aLen2b/aLen2
      R(2,3) = aLen2c/aLen2
c
      a_J11 = (-0.25*(1-c_s))*x1 + (0.25*(1-c_s))*x2 + (0.25*(1+c_s))*x3 +
     & (-0.25*(1+c_s))*x4
      a_J12 = (-0.25*(1-c_s))*z1 + (0.25*(1-c_s))*z2 + (0.25*(1+c_s))*z3 +
     & (-0.25*(1+c_s))*z4
      a_J21 = (-0.25*(1-c_r))*x1 + (-0.25*(1+c_r))*x2 + (0.25*(1+c_r))*x3 +
     & (0.25*(1-c_r))*x4
      a_J22 = (-0.25*(1-c_r))*z1 + (-0.25*(1+c_r))*z2 + (0.25*(1+c_r))*z3 +
     & (0.25*(1-c_r))*z4
c
      b_J11 = (-0.25*(1-c_s))*x1 + (0.25*(1-c_s))*x2 + (0.25*(1+c_s))*x3 +
     & (-0.25*(1+c_s))*x4
      b_J12 = (-0.25*(1-c_s))*y1 + (0.25*(1-c_s))*y2 + (0.25*(1+c_s))*y3 +
     & (-0.25*(1+c_s))*y4
      b_J21 = (-0.25*(1-c_r))*x1 + (-0.25*(1+c_r))*x2 + (0.25*(1+c_r))*x3 +
     & (0.25*(1-c_r))*x4
      b_J22 = (-0.25*(1-c_r))*y1 + (-0.25*(1+c_r))*y2 + (0.25*(1+c_r))*y3 +
     & (0.25*(1-c_r))*y4
c
      c_J11 = (-0.25*(1-c_s))*y1 + (0.25*(1-c_s))*y2 + (0.25*(1+c_s))*y3 +
     & (-0.25*(1+c_s))*y4
      c_J12 = (-0.25*(1-c_s))*z1 + (0.25*(1-c_s))*z2 + (0.25*(1+c_s))*z3 +
     & (-0.25*(1+c_s))*z4
      c_J21 = (-0.25*(1-c_r))*y1 + (-0.25*(1+c_r))*y2 + (0.25*(1+c_r))*y3 +
     & (0.25*(1-c_r))*y4
      c_J22 = (-0.25*(1-c_r))*z1 + (-0.25*(1+c_r))*z2 + (0.25*(1+c_r))*z3 +
     & (0.25*(1-c_r))*z4
c
c
      a_Jacob1 = (a_J11*a_J22 - a_J12*a_J21)
      a_Jacob2 = (b_J11*b_J22 - b_J12*b_J21)
      a_Jacob3 = (c_J11*c_J22 - c_J12*c_J21)
c
      a_Jacob = sqrt(a_Jacob1**2.0d00 + a_Jacob2**2.0d00 + a_Jacob3**2.0d00)
c
c======================================================================
      num=nnode
c
      do i = 1, num
         dum=3.0*(i-1.0)
         Transformation_M(dum+1,dum+1)=R(1,1)
         Transformation_M(dum+1,dum+2)=R(1,2) 
         Transformation_M(dum+1,dum+3)=R(1,3)
         Transformation_M(dum+2,dum+1)=R(2,1)
         Transformation_M(dum+2,dum+2)=R(2,2)
         Transformation_M(dum+2,dum+3)=R(2,3)
         Transformation_M(dum+3,dum+1)=R(3,1)
         Transformation_M(dum+3,dum+2)=R(3,2)
         Transformation_M(dum+3,dum+3)=R(3,3)
      end do
c
      call k_matrix_transpose(Transformation_M,Transformation_M_T,
     $ ndofel,ndofel)
c
c
      do i = 1, nnode
         coord_l(1,i)=(R(1,1)*co_de(1,i)+R(1,2)*co_de(2,i)
     & +R(1,3)*co_de(3,i))
         coord_l(2,i)=(R(2,1)*co_de(1,i)+R(2,2)*co_de(2,i)
     & +R(2,3)*co_de(3,i))
         coord_l(3,i)=(R(3,1)*co_de(1,i)+R(3,2)*co_de(2,i)
     & +R(3,3)*co_de(3,i))
      end do
c    
      return
      end
c======================================================================
      subroutine k_shape_fun(i,sf)
      INCLUDE 'ABA_PARAM.INC'
      dimension sf(4), GP_coord(2)
c
      if (i .eq. 1) then
         GP_coord(1)=-sqrt(1.0d0/3.0d0)
         GP_coord(2)=-sqrt(1.0d0/3.0d0)
      elseif (i .eq. 2) then
         GP_coord(1)= sqrt(1.0d0/3.0d0)
         GP_coord(2)=-sqrt(1.0d0/3.0d0)
      elseif (i .eq. 3) then
         GP_coord(1)= sqrt(1.0d0/3.0d0)
         GP_coord(2)= sqrt(1.0d0/3.0d0)
      elseif (i .eq. 4) then
         GP_coord(1)=-sqrt(1.0d0/3.0d0)
         GP_coord(2)= sqrt(1.0d0/3.0d0)
      end if
c
      sf(1)=(1-GP_coord(1))*(1-GP_coord(2))*0.25
      sf(2)=(1+GP_coord(1))*(1-GP_coord(2))*0.25
      sf(3)=(1+GP_coord(1))*(1+GP_coord(2))*0.25
      sf(4)=(1-GP_coord(1))*(1+GP_coord(2))*0.25
c
      return
      end
c======================================================================
      subroutine k_matrix_multiply(A,B,C,l,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(l,n),B(n,m),C(l,m)
c
      call k_matrix_zero(C,l,m)
c
      do i = 1, l
         do j = 1, m
            do k = 1, n
               C(i,j)=C(i,j)+A(i,k)*B(k,j)
            end do
         end do
      end do
c
      return
      end
c======================================================================
      subroutine k_matrix_plus_scalar(A,B,c,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n,m),B(n,m)
c
      do i = 1, n
         do j = 1, m
            A(i,j)=A(i,j)+c*B(i,j)
         end do
      end do
c
      return
      end
c======================================================================
      subroutine k_matrix_transpose(A,B,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n,m),B(m,n)
c
      do i = 1, n
         do j = 1, m
            B(j,i)=A(i,j)
         end do
      end do
c
      return
      end
c======================================================================
      subroutine k_matrix_zero(A,n,m)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n,m)
c
      do i = 1, n
         do j = 1, m
            A(i,j)=0.d0
         end do
      end do
c
      return
      end
c======================================================================
      subroutine k_vector_zero(A,n)
      INCLUDE 'ABA_PARAM.INC'
      dimension A(n)
c
      do i = 1, n
         A(i)=0.d0
      end do
c
      return
      end
c======================================================================
      subroutine k_Mac(pM,a,b)
      INCLUDE 'ABA_PARAM.INC'
c
      if ((a-b) .GE. 0.0) then
         pM=a-b
      elseif ((a-b) .LT. 0.0) then
         pM=0.d0
      end if
c
      return
      end
c======================================================================
c=================================END==================================
c======================================================================









      
